import gym
from gym import spaces
from universe import error

class Env(gym.Env):
    metadata = {
        'runtime.vectorized': True
    }

    # User should set this!
    n = None

    @property
    def monitor(self):
        if not self.metadata['runtime.vectorized']:
            # Just delegate if we're not actually vectorized (like
            # Unvectorize)
            return super(Env, self).monitor

        if not hasattr(self, '_monitor'):
            # Not much we can do if we don't know how wide we'll
            # be. This can happen when closing.
            if self.n is None:
                raise error.Error('You must call "configure()" before accesssing the monitor for {}'.format(self))

            # Circular dependencies :(
            from universe import wrappers
            from universe.vectorized import monitoring
            # We need to maintain pointers to these to them being
            # GC'd. They have a weak reference to us to avoid cycles.
            self._unvectorized = [wrappers.WeakUnvectorize(self, i) for i in range(self.n)]
            # Store reference to avoid GC
            # self._render_cached = monitoring.RenderCache(self)
            self._monitor = monitoring.Monitor(self._unvectorized)
        return self._monitor

class Wrapper(Env, gym.Wrapper):
    autovectorize = True
    standalone = True

    def __init__(self, env=None):
        super(Wrapper, self).__init__(env)
        if env is not None and not env.metadata.get('runtime.vectorized'):
            if self.autovectorize:
                # Circular dependency :(
                from universe import wrappers
                env = wrappers.Vectorize(env)
            else:
                raise error.Error('This wrapper can only wrap vectorized envs (i.e. where env.metadata["runtime.vectorized"] = True), not {}. Set "self.autovectorize = True" to automatically add a Vectorize wrapper.'.format(env))

        if env is None and not self.standalone:
            raise error.Error('This env requires a non-None env to be passed. Set "self.standalone = True" to allow env to be omitted or None.')

        self.env = env

    def _configure(self, **kwargs):
        super(Wrapper, self)._configure(**kwargs)
        assert self.env.n is not None, "Did not set self.env.n: self.n={} self.env={} self={}".format(self.env.n, self.env, self)
        self.n = self.env.n

class ObservationWrapper(Wrapper, gym.ObservationWrapper):
    pass

class RewardWrapper(Wrapper, gym.RewardWrapper):
    pass

class ActionWrapper(Wrapper, gym.ActionWrapper):
    pass



import os
import random
import re
import signal
import time

from universe import error
from universe.twisty import reactor
from twisted.internet import defer, protocol, task
import twisted.internet.error
import logging

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

class ConnectionTimer(protocol.Protocol):
    def connectionMade(self):
        self.transport.loseConnection()

def connection_timer_factory():
    factory = protocol.ClientFactory()
    factory.protocol = ConnectionTimer
    return factory

class StopWatch(object):
    def start(self):
        self.start_time = time.time()

    def stop(self):
        return time.time() - self.start_time

# TODO: clean this up
def start(endpoint, max_attempts=0):
    # Use an object for timing so that we can mutate it within the closure
    stop_watch = StopWatch()

    def success(client):
        return stop_watch.stop()

    def error(failure, retry):
        # some websocket implementations (like websocketcpp) can fail when connections are lost too quickly
        if retry == 0:
            raise ConnectionTimerException("Max retries")
        backoff = 2 ** (max_attempts - retry + 1) + random.randint(42, 100)
        logger.info('Throttling down websocket creation after connection error (this is normal) - waiting %dms - '
                    'error details %s', backoff, error)
        d = task.deferLater(reactor, backoff / 1000., go, retry - 1)
        return d

    def go(retry):
        stop_watch.start()
        factory = connection_timer_factory()
        d = endpoint.connect(factory)
        d.addCallback(success)
        d.addErrback(error, retry)
        return d

    return go(retry=max_attempts)

def measure_clock_skew(label, host):
    cmd = ['ntpdate', '-q', '-p', '8', host]
    extra_logger.info('[%s] Starting network calibration with %s', label, ' '.join(cmd))
    skew = Clockskew(label, cmd)
    # TODO: search PATH for this?
    process = reactor.spawnProcess(skew, '/usr/sbin/ntpdate', cmd, {})
    # process = reactor.spawnProcess(skew, '/bin/sleep', ['sleep', '2'], {})

    t = float(os.environ.get('UNIVERSE_NTPDATE_TIMEOUT', 20))
    def timeout():
        if process.pid:
            logger.error('[%s] %s call timed out after %ss; killing the subprocess. This is ok, but you could have more accurate timings by enabling UDP port 123 traffic to your env. (Alternatively, you can try increasing the timeout by setting environment variable UNIVERSE_NTPDATE_TIMEOUT=10.)', label, ' '.join(cmd), t)
            process.signalProcess(signal.SIGKILL)
            process.reapProcess()
    # TODO: make this part of the connection string
    reactor.callLater(t, timeout)
    return skew.deferred

class Clockskew(protocol.ProcessProtocol):
    def __init__(self, label, cmd):
        self.label = label
        self._cmd = cmd

        self.deferred = defer.Deferred()
        self.out = []
        self.err = []

    def outReceived(self, data):
        self.out.append(data)

    def errReceived(self, data):
        self.err.append(data)

    def processExited(self, reason):
        if isinstance(reason.value, twisted.internet.error.ProcessDone):
            out = b''.join(self.out).decode('utf-8')
            match = re.search('offset ([\d.-]+) sec', out)
            if match is not None:
                offset = float(match.group(1))
                self.deferred.callback(offset)
            else:
                self.deferred.errback(error.Error('Could not parse offset: %s', out))
        else:
            err = b''.join(self.err)
            self.deferred.errback(error.Error('{} failed with status {}: stderr={!r}'.format(self._cmd, reason.value.exitCode, err)))

class ConnectionTimerException(Exception):
    pass



import weakref

from gym import monitoring

class Monitor(object):
    def __init__(self, env_n):
        self.monitor_n = [monitoring.Monitor(env) for env in env_n]

    @property
    def env(self):
        # The real env is the first unwrapped env. Maybe we should
        # maintain our own weakref rather than doing this.
        return self.monitor_n[0].env.env

    def start(self, directory, video_callable=None, seed_n=None, force=False,
              resume=False, write_upon_reset=False, uid=None):
        if seed_n is None:
            seed_n = [None] * len(self.monitor_n)
        # There's way to seed just one of the vectorized environments,
        # so we have to do the seeding ourselves outside of the
        # underlying monitor instances.
        #
        # The monitor will call the .seed method on the
        # WeakUnvectorized env, which just returns rather than
        # actually re-seeding the env.
        self.env.seed(seed_n)

        for i, monitor in enumerate(self.monitor_n):
            # Only allow recording of video in first monitor
            if i > 0:
                video_callable = False
            # Seed gets passed in but just recorded, not used.
            monitor.start(directory=directory, video_callable=video_callable,
                          force=force, resume=resume, write_upon_reset=write_upon_reset, uid=uid)

    def close(self, *args, **kwargs):
        [monitor.close(*args, **kwargs) for monitor in self.monitor_n]

    def _before_reset(self):
        return [monitor._before_reset() for monitor in self.monitor_n]

    def _after_reset(self, observation_n):
        return [monitor._after_reset(observation) for monitor, observation in zip(self.monitor_n, observation_n)]

    def _before_step(self, action_n):
        return [monitor._before_step(action) for monitor, action in zip(self.monitor_n, action_n)]

    def _after_step(self, observation_n, reward_n, done_n, info):
        return [monitor._after_step(o, r, d, i) for monitor, o, r, d, i in zip(self.monitor_n, observation_n, reward_n, done_n, info['n'])]



import gym
from universe import envs, spaces
from universe.wrappers import gym_core_sync

from universe.wrappers.action_space import SafeActionSpace
from universe.wrappers.gym_core import GymCoreAction, GymCoreObservation, CropAtari
from universe.wrappers.blocking_reset import BlockingReset
from universe.wrappers.diagnostics import Diagnostics
from universe.wrappers.joint import Joint
from universe.wrappers.logger import Logger
from universe.wrappers.multiprocessing_env import WrappedMultiprocessingEnv, EpisodeID
from universe.wrappers.render import Render
from universe.wrappers.throttle import Throttle
from universe.wrappers.timer import Timer
from universe.wrappers.vectorize import Vectorize, Unvectorize, WeakUnvectorize
from universe.wrappers.vision import Vision

def wrap(env):
    return Timer(Render(Throttle(env)))

def WrappedVNCEnv():
    return wrap(envs.VNCEnv())

def WrappedGymCoreEnv(gym_core_id, fps=None, rewarder_observation=False):
    # Don't need to store the ID on the instance; it'll be retrieved
    # directly from the spec
    env = wrap(envs.VNCEnv(fps=fps))
    if rewarder_observation:
        env = GymCoreObservation(env, gym_core_id=gym_core_id)
    return env

def WrappedGymCoreSyncEnv(gym_core_id, fps=60, rewarder_observation=False):
    spec = gym.spec(gym_core_id)
    env = gym_core_sync.GymCoreSync(BlockingReset(wrap(envs.VNCEnv(fps=fps))))
    if rewarder_observation:
        env = GymCoreObservation(env, gym_core_id=gym_core_id)
    elif spec._entry_point.startswith('gym.envs.atari:'):
        env = CropAtari(env)

    return env

def WrappedFlashgamesEnv():
    keysym = spaces.KeyEvent.by_name('`').key
    return wrap(envs.VNCEnv(probe_key=keysym))

def WrappedInternetEnv(*args, **kwargs):
    return wrap(envs.InternetEnv(*args, **kwargs))

def WrappedStarCraftEnv(*args, **kwargs):
    return wrap(envs.StarCraftEnv(*args, **kwargs))

def WrappedGTAVEnv(*args, **kwargs):
    return wrap(envs.GTAVEnv(*args, **kwargs))

def WrappedWorldOfGooEnv(*args, **kwargs):
    return wrap(envs.WorldOfGooEnv(*args, **kwargs))



from gym.benchmarks import register_benchmark

register_benchmark(
    id='Atari7VNC-v0',
    score_method='average_last_100_episodes',
    tasks={
        "BeamRider": {
            "env_id": "VNCBeamRider-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "Breakout": {
            "env_id": "VNCBreakout-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "Enduro": {
            "env_id": "VNCEnduro-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "Pong": {
            "env_id": "gym-core.Pong-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "Qbert": {
            "env_id": "VNCQbert-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "Seaquest": {
            "env_id": "VNCSeaquest-v3",
            "seeds": 1,
            "timesteps": 10000000
        },
        "SpaceInvaders": {
            "env_id": "VNCSpaceInvaders-v3",
            "seeds": 1,
            "timesteps": 10000000
        }
    })
